package cz.upol.wall;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import javafx.beans.property.BooleanProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleIntegerProperty;

/**
 * Trida reprezentuje vnitrni stav hry
 */
public class GameState {
	
	/** pocet sloupcu cihel */
	public static final int BRICK_COLS = 10;
	
	/** pocet rad cihel */
	public static final int BRICK_ROWS = 5;
	
	/** sirka cihly */
	public static final int BRICK_WIDTH = 4;
	
	/** pocet radku herni plochy vcetne cihel */
	public static final int ROWS = 20;
	
	/** pocet sloupcu herni plochy */
	public static final int COLUMNS = BRICK_COLS * BRICK_WIDTH;
	
	/** velikost hrace (palky) */
	public static final int PADDLE_WIDTH =  5;
	
	/** pozice palky */
	public static final int PADDLE_ROW = 20;
	
	/** seznam vsech cihel */
	private List<Brick> bricks;
	
	/** mic */
	private Ball ball = new Ball(0, 0);

	/** pozice hrace, resp. sloupec, na kterem se nachazi zacatek hrace (palky) */
	private IntegerProperty paddlePos = new SimpleIntegerProperty(0);
	
	/** ziskane body */
	private IntegerProperty score = new SimpleIntegerProperty(0);
	
	/** priznak, zda je hra aktivni */
	private BooleanProperty active = new SimpleBooleanProperty(false);
	
	/**
	 * Vytvori hru a inicializuje hodnoty tak,
	 * aby odpovidaly zacatku hry 
	 */
	public GameState() {
		// inicializace cihel
		bricks = new ArrayList<>();
		for (int r = 0; r < BRICK_ROWS; r++)
			for (int c = 0; c < BRICK_COLS; c++)
				bricks.add(new Brick(r, c));

		setDefaultValues();
	}
	
	/**
	 * Reinicializuje stav hry, tak jak ma byt na zacatku hry
	 */
	public void reinitialize() {
		bricks.forEach(b -> b.setActive(true));
		setDefaultValues();
	}
	
	/**
	 * Inicializuje umisteni mice, hrace a stav hry
	 */
	private void setDefaultValues() {
		ball.setColumn(COLUMNS / 2);
		ball.setRow(PADDLE_ROW - 1);
		ball.initializeDirection();
		
		paddlePos.setValue((COLUMNS - PADDLE_WIDTH) / 2);
		
		score.setValue(0);
		active.setValue(false);
	}

	/**
	 * Vraci seznam vsech cihel. Jelikoz odkazujeme primo
	 * na vnitrni seznam, pro jistotu jej udelame nemodifikovatelny.
	 */
	public List<Brick> getBricks() {
		return Collections.unmodifiableList(bricks);
	}

	/**
	 * @return vraci mic
	 */
	public Ball getBall() {
		return ball;
	}
	
	/**
	 * Aktualizuje hru, posune mic, resi pripadne kolize
	 * @return Vraci true, pokud hra muze pokracovat, jinak false.
	 */
	public boolean update() {
		ball.update();
		return checkBall(ball);
	}
	
	/**
	 * Resi kolize mice
	 * @return  Vraci true, pokud hra muze pokracovat, jinak false.
	 */
	private boolean checkBall(Ball ball) {
		
		int r = ball.getRow();
		int c = ball.getColumn();

		// naraz do svisleho okraje
		if ((c == 0) || (c == COLUMNS - 1)) ball.hBounce();
		
		// testuje, jestli nenarazil do cihly
		// (reseni neni prilis efektivni, ale dostacuje)
		Optional<Brick> collision = bricks.stream().filter(b -> b.isActive() && (b.getColumn() == c / BRICK_WIDTH) && (b.getRow() == r)).findFirst();
		if (collision.isPresent()) {
			ball.vBounce();                       // odraz
			collision.get().setActive(false);     // cihla je odstranena
			score.setValue(score.getValue() + 1); // navyseni score
		} else {
			int paddleCol = paddlePos.getValue();
			// odraz od horni stany
			if (r == 0) ball.vBounce();
			
			// odraz od uzivatele (palky)
			if ((r == PADDLE_ROW - 1) && (c >= paddleCol) && (c < paddleCol + PADDLE_WIDTH)) ball.vBounce();
			
			// propadnuti mice pres spodni okraj
			if (r == ROWS) {
				setActive(false);
				return false;
			}
		}
		
		return true;
	}

	/**
	 * Posune hrace doleva
	 */
	public void moveLeft() {
		if (!isActive()) return;
		int paddleCol = paddlePos.getValue();
		if (paddleCol > 0) paddlePos.setValue(paddleCol - 1);
	}

	/**
	 * Posune hrace doprava
	 */
	public void moveRight() {
		if (!isActive()) return;
		int paddleCol = paddlePos.getValue();
		if (paddleCol + PADDLE_WIDTH < COLUMNS) paddlePos.setValue(paddleCol + 1);
	}
	
	//
	// Vlastnosti pouzivane JavaFX
	//

	public final IntegerProperty paddlePosProperty() {
		return this.paddlePos;
	}

	public final IntegerProperty scoreProperty() {
		return this.score;
	}
	
	public final int getScore() {
		return this.scoreProperty().get();
	}
	
	public final void setScore(final int score) {
		this.scoreProperty().set(score);
	}

	public final BooleanProperty activeProperty() {
		return this.active;
	}
	
	public final boolean isActive() {
		return this.activeProperty().get();
	}
	
	public final void setActive(final boolean active) {
		this.activeProperty().set(active);
	}
}
