package cz.upol.wall;

import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.KeyValue;
import javafx.animation.Timeline;
import javafx.application.Application;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.effect.DropShadow;
import javafx.scene.effect.InnerShadow;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.scene.layout.StackPane;
import javafx.scene.paint.Color;
import javafx.scene.shape.Ellipse;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.Text;
import javafx.stage.Stage;
import javafx.util.Duration;

/**
 * Graficka reprezentace hry.
 * 
 * Veskera logika hry je umistena v {@link GameState}
 */
public class WallApp extends Application {
	
	/** 
	 * cas (v milisekundach) mezi jednotlivymi aktualizacemi hry;
	 * cim mensi, tim rychlejsi
	 */
	private static final int DELAY = 100; // ms
	
	/** velikost odpovidajici jednomu sloupci/radku v GameState */
	private static final int UNIT_SIZE = 15;
	
	/** stav hry */
	private GameState game = new GameState();
	
	/** herni plocha */
	private Pane gamePane;
	
	/** textovy popisek se ziskanymi body */
	private Text lbScore;
	
	/** tlacitko start */ 
	private Button btnStart;
	
	/** zdroj casu pro aktualizaci hry */
	private Timeline timeline;
	
	
	@Override
	public void start(Stage primaryStage) throws Exception {
		
		createGamePane();
		
		lbScore = new Text();
		lbScore.textProperty().bind(game.scoreProperty().asString());
		
		timeline = new Timeline();
		KeyFrame updates = new KeyFrame(
		        Duration.millis(DELAY),
		        e -> { if (!game.update()) gameOver(); }
		        );
		timeline.getKeyFrames().add(updates);
		timeline.setCycleCount(Animation.INDEFINITE);
		
		btnStart = new Button("Start");
		btnStart.visibleProperty().bind(game.activeProperty().not());
		btnStart.setOnAction(e -> {
			game.reinitialize();
			game.setActive(true);
			gamePane.setOpacity(1);
			timeline.play();
		});
		
		btnStart.setScaleX(2);
		btnStart.setScaleY(2);
		
		// pripadna transformace pro effekt ;-]
		// btnStart.setRotate(10);
		
		BorderPane root = new BorderPane();
		
		StackPane gameStack = new StackPane();
		gameStack.getChildren().addAll(gamePane, btnStart);
		
		HBox scorePane = new HBox(10, new Text("Score: "), lbScore);
		
		root.setCenter(gameStack);
		root.setTop(scorePane);
		
		Scene scene = new Scene(root, 600, 500);
		scene.setOnKeyPressed(this::dispatchKeyEvents);
		primaryStage.setTitle("Wall");
		primaryStage.setScene(scene);
		primaryStage.show();
	}
	
	/**
	 * Vytvori herni plochu a provaze ji se stavem hry
	 */
	private void createGamePane() {
		gamePane = new Pane();
		gamePane.setMaxWidth(UNIT_SIZE * GameState.COLUMNS);
		gamePane.setMaxHeight(UNIT_SIZE * GameState.ROWS);
		
		for (Brick brick : game.getBricks())
			gamePane.getChildren().add(createBrick(brick));
		
		gamePane.getChildren().addAll(
				createBall(game.getBall()),
				createPaddle());
	}
	
	/**
	 * Vytvori obdelnik reprezentujici cihlu a navaze jej na odpovidajici cihlu
	 * ve stavu hry
	 */
	private Rectangle createBrick(Brick brick) {
		Rectangle rect = new Rectangle(UNIT_SIZE  * GameState.BRICK_WIDTH - 1, UNIT_SIZE - 1);
		
		if ((brick.getRow() + brick.getColumn()) % 2 == 0) rect.setFill(Color.DARKRED);
		else rect.setFill(Color.RED);
		
		// provazani vlastnosti
		rect.xProperty().bind(brick.columnProperty().multiply(UNIT_SIZE * GameState.BRICK_WIDTH));
		rect.yProperty().bind(brick.rowProperty().multiply(UNIT_SIZE));
		rect.visibleProperty().bind(brick.activeProperty());
		
		// pripadny efekt - vpadly stin
		InnerShadow is = new InnerShadow();
		is.setOffsetY(1.0);
		is.setOffsetX(1.0);
		is.setColor(Color.GRAY);
		rect.setEffect(is);
		
		return rect;
	}
	
	/**
	 * Vytvori mic a navaze jej na stav hry
	 */
	private Ellipse createBall(Ball ball) {
		
		Ellipse ballShape = new Ellipse(UNIT_SIZE / 2, UNIT_SIZE / 2 - 1);
		ballShape.centerXProperty().bind(ball.columnProperty().multiply(UNIT_SIZE).add(UNIT_SIZE / 2));
		ballShape.centerYProperty().bind(ball.rowProperty().multiply(UNIT_SIZE).add(UNIT_SIZE / 2));
		
		// pripadna aplikace transformace
		Timeline ballTimeline = new Timeline();
		ballTimeline.getKeyFrames().add(new KeyFrame(Duration.millis(500), new KeyValue(ballShape.rotateProperty(), 360)));
		ballTimeline.setCycleCount(Timeline.INDEFINITE);
		ballTimeline.play();
		
		return ballShape;
	}
	
	
	/**
	 * Vytvori hrace (palku) a navaze ji na stav hry
	 */
	private Rectangle createPaddle() {
		Rectangle paddle = new Rectangle(UNIT_SIZE * GameState.PADDLE_WIDTH, UNIT_SIZE);
		paddle.xProperty().bind(game.paddlePosProperty().multiply(UNIT_SIZE));
		paddle.setY(UNIT_SIZE * GameState.PADDLE_ROW);
		
		// pripadny effekt - stin pod hracem
		DropShadow ds = new DropShadow();
		ds.setOffsetY(3.0);
		ds.setOffsetX(3.0);
		ds.setColor(Color.GRAY);
				
		paddle.setEffect(ds);
		return paddle;
	}
	
	
	/**
	 * Provede zastaveni hry
	 */
	private void gameOver() {
		timeline.stop();
		gamePane.setOpacity(0.2);
	}
	
	/**
	 * Sleduje stisknete klavesy a pradava udalosti do stavu hry
	 */
	private void dispatchKeyEvents(KeyEvent e) {
		switch (e.getCode()) {
		case LEFT: game.moveLeft(); break;
		case RIGHT: game.moveRight(); break;
		default:
		}
	}
	
	public static void main(String[] args) {
		launch(args);
	}

}
